<domain type='kvm'>
    <name>{{ name }}</name>
    <memory unit='MiB'>{{ mem_size }}</memory>
    <iothreads>1</iothreads>
    <pm>
      <suspend-to-disk enabled='no'/>
      <suspend-to-mem enabled='no'/>
    </pm>
    <os>
      <type arch='x86_64' machine='pc'>hvm</type>
      <bootmenu enable='no'/>
      <bios useserial='no'/>
    </os>
    <features>
      <acpi/>
      <apic/>
      <pae/>
      <vmport state='off'/>
    </features>
    <devices>
      <emulator>{{ qemu_kvm }}</emulator>
      <memballoon model='none'/>
      <controller type='usb' model='none'>
      </controller>
      <disk type='file' device='disk'>
        <driver name='qemu' type='qcow2'/>
        <source file='DISK_PATH'/>
        <target dev='DISK_DEV' bus='virtio'/>
        <boot order='1'/>
      </disk>
      <channel type='unix'>
        <source mode='bind'/>
        <target type='virtio' name='org.qemu.guest_agent.0'/>
      </channel>
      <rng model='virtio'>
        <backend model='random'>{{ '/dev/urandom' if libvirt_ver >= 2002001 else '/dev/random' }}</backend>
        <address type='pci' domain='0x0000' bus='0x00' slot='0x09' function='0x0'/>
      </rng>
      <console type='pty'>
        <target type='virtio' port='0'/>
      </console>
      <video>
        <model type='cirrus' vram='16384' heads='1'/>
        <alias name='video0'/>
        <address type='pci' domain='0x0000' bus='0x00' slot='0x02' function='0x0'/>
      </video>
    </devices>
    <clock offset='utc'>
      <timer name='kvmclock'>
      </timer>
    </clock>
</domain>
