{% import 'sysprep-macros.j2' as macros %}
{% include 'sysprep-base.j2' %}

{{ macros.iscsi(name=iscsi_name, hostname=hostname) }}
{{ macros.network_devices_el(mappings=mappings) }}

edit /boot/grub2/grub.cfg:s/set timeout=5/set timeout=0/g

delete /etc/systemd/system/multi-user.target.wants/kdump.service
delete /etc/systemd/system/multi-user.target.wants/postfix.service
